/*****************************************************************************
 Copyright (c) 1993 UUNET Technologies
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
 3. All advertising materials mentioning features or use of this software
    must display the following acknowledgement:
      This product includes software developed by UUNET Technologies
 4. The name UUNET Technologies may not be used to endorse or promote products  
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE UUNET Technologies ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ARE DISCLAIMED.  IN NO EVENT SHALL THE UUNET Technologies BE LIABLE
 FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

*****************************************************************************/


#include <sys/types.h>
#include <sfio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <assert.h>
#include <malloc.h>
#include <string.h>

#include <general.h>

#include <rsaref.h>
#include <r_encode.h>

#include <keys.h>
#include <key_public.h>

#include <muse.h>

#include <timestuff.h>

int Nntp_port = 119;

char *nntpserver = "";
int socketfd;
Sfile_t *socketfile;

char errorstr[1024];

void initmulti(), *getmulti();

char *slurpnntpresponse (Sfio_t *);

const char *service = "nntp";

int main(argc, argv)
     int argc;
     char *argv[];
{
  char *response, *msgidloc, *foo;
  char msgid[256];
  int msgidlen, K;
  void *data;
  mpacket packet;
  mpacket *pck = &packet;
  int pcksize;
  int verificationstatus, operationstatus;
  int artlen;
  Sfio_t *in, *out;
  void *sig;
  char base64[1024];
  l_fp packettime, receivedtime, verifiedtime, donetime, temptime;
  struct timeval tv;

  sfioinit();
  sfset(sfout,SF_LINE,1);

  readkeystatus();

  nntpserver = NNTPHOST;
  socketfd = get_tcp_conn(nntpserver,service);
  if (socketfd < 0) {
    sfprintf(sferr, "Bad return from get_tcp_conn(%s,%s): %d", 
	     nntpserver, service, socketfd);
    return 4;
  }
  socketfile = sfnew(NULL,NULL,-1,socketfd, SF_READ|SF_WRITE);

  response = slurpnntpresponse(socketfile);
  if (strncmp(response, "200", 3) != 0) {
    sfprintf(sferr, "Bad return code from open: %3s\n", response);
    return 3;
  }

  initmulti();

  while (data = getmulti(&pcksize)) {
    gettimeofday(&tv, NULL);
    TVTOTS(&tv,&receivedtime);

    raw2mpacket(data, pcksize, pck, &artlen, &sig);
    packettime.l_ui = pck->header.sec;
    packettime.l_uf = pck->header.fsec;
    in = sfnew(NULL, sig ,sizeof(pck->header.rsa_sig),
	       0,SF_STRING|SF_READ);
    out = sfnew(NULL, base64,sizeof(base64),
		0,SF_STRING|SF_WRITE);
    encode_base64(in,out);
    sfclose(in);
    sfclose(out);
    
    bzero(sig, sizeof(pck->header.rsa_sig));
    K = pcksize;
    if (getkeystatus(pck->header.keyno) & Krevoked) {
      if (pck->header.cmdfkr == Crevoke) {
	continue; /* ignore revoke messages for already revoked keys */
      } else {
	strcpy(msgid, "Ekeyrevoked");
	operationstatus = 1;
	verificationstatus = -1;
	sfsprintf(errorstr, sizeof(errorstr), "key already revoked, cmd=%d",
		 pck->header.keyno);
      }
    } else {
      verificationstatus = 
	R_VerifyPEMSignature(data, &K, data, K,
			     base64, strlen(base64),
			     0, DA_MD5,
			     &PUBLIC_KEYS[pck->header.keyno]);
      gettimeofday(&tv, NULL);
      TVTOTS(&tv,&verifiedtime);
      if (verificationstatus != 0)
	{
	  RSAerr(verificationstatus);
	  setkeystatus(pck->header.keyno, getkeystatus(pck->header.keyno) |
		       Kfailed);
      } else {
	setkeystatus(pck->header.keyno, getkeystatus(pck->header.keyno) |
		     Kused);
      }
      switch (pck->header.cmdfkr) {
      case Cmessage:
	pck->message[artlen] = '\0';
	msgidloc = ((char *) (&pck->message)) + pck->header.mesgid_off;
	if (foo = strchr(msgidloc, '>')) {
	  msgidlen = foo - msgidloc + 1;
	  strncpy(msgid,msgidloc,msgidlen);
	  msgid[msgidlen] = 0;
	} else {
	  strcpy(msgid,"NONE");
	}
	operationstatus = nntpofferfkr(socketfile, msgid, &pck->message);
	break;
      case Crevoke:
	strcpy(msgid, "Crevoke");
	setkeystatus(pck->header.keyno, getkeystatus(pck->header.keyno) |
		     Krevoked);
	operationstatus = writekeystatus();
	sfsprintf(errorstr, sizeof(errorstr), "%d %d",
		  operationstatus, errno);
	break;
      case Cnoop:
	strcpy(msgid, "Cnoop");
	operationstatus = 0;
	strcpy(errorstr, "success");
	break;
      case Cversion:
	strcpy(msgid, "Cversion");
	operationstatus = 1;
	strcpy(errorstr, "not implemented");
	break;
      case Cstatistics:
	strcpy(msgid, "Cstatistics");
	operationstatus = 1;
	strcpy(errorstr, "not implemented");
      default:
	sfsprintf(msgid, sizeof(msgid), "C#%d", pck->header.cmdfkr);
	  operationstatus = 1;
	sfsprintf(errorstr, sizeof(errorstr), "unknown command code %d",
		  pck->header.cmdfkr);
      }
    }
    gettimeofday(&tv, NULL);
    TVTOTS(&tv,&donetime);
    temptime = donetime;
    M_SUB(temptime.l_ui, temptime.l_uf,packettime.l_ui, packettime.l_uf);
    sfprintf(sfout,
	     "%s %d %d %s %s %d %d <%s>\n",
	     msgid,
	     pck->header.sequence,
	     pck->header.keyno,
	     ulfptoa(receivedtime, 5),
	     ulfptoa(temptime, 5),
	     verificationstatus,
	     operationstatus,
	     errorstr);
    
    if ((pck->header.sequence & 0xFF) == 0) {
      writekeystatus(); /* write the key status every 256 articles, even
			   if no keys have been revoked */
    }
    
  } /* end of main while */

  writekeystatus();

  sfprintf(socketfile, "quit\r\n");
  sfclose(socketfile);
}  

char *slurpnntpresponse(socket)
     Sfio_t *socket;
{
  char *r, *last;
  int foo;
  static char resbuf[1024];

  r = sfgetr(socket, '\n',1);
  strncpy(resbuf, r, 1023);
  resbuf[1023] = 0;
  last = resbuf + strlen(resbuf) -1;
  if ((*last) == '\r') {
    *last = 0;
  }
  return resbuf;
}

int nntpofferfkr(socket,msgid,message)
     Sfio_t *socket;
     char *message, *msgid;
{
  char *response;
  int done;
  int retval;

  sfprintf(socket, "ihave %s\r\n", msgid); 

  response = slurpnntpresponse(socket);
  if (strncmp(response, "435", 3) == 0) {
    sprintf(errorstr,"server didn\'t want article: %s", response);
    return 2;
  } else if (strncmp(response, "335", 3) != 0) {
    sprintf(errorstr,"bad response to IHAVE: %s", response);
    return 0;
  }



  for (; *message != 0; )
    {
      if (*message == '.') {
	sfputc(socket,'.');
      }
      for (; *message != '\n'; message++)
	sfputc(socket,(*message));
      sfputc(socket,(*message));
      message++;
    }
  sfputr(socket,".\r\n",-1);
/*  sfsync(socket); */

  response = slurpnntpresponse(socket);

  if (strncmp(response,"235",3) != 0) {
    sprintf(errorstr,"bad response to data: %s", response);
    return 0;
  }

  sprintf(errorstr,"article successfully offered: %s",response);
  return 1;
}



#define DLEN (1024 * 64)

#ifndef IP_ADD_MEMBERSHIP

#define IP_ADD_MEMBERSHIP 5

/* I wonder what file this belongs in? */
struct ip_mreq {
	struct in_addr imr_multiaddr;   /* multicast group to join */
	struct in_addr imr_interface;   /* interface to join on    */
};

#endif	/* IP_ADD_MEMBERSHIP */

u_long makeaddr(int g1, int g2, int g3, int g4)
{
	return htonl((g1<<24) | (g2<<16) | (g3<<8) | g4);
}

struct sockaddr_in multiaddr;
struct ip_mreq mreq;
int multisock;
int fromlen;
char *multidata;
int one = 1;

void initmulti ()
{
  int buflen = 65536;
  int retstat = 0;

  multiaddr.sin_family = AF_INET;
  multiaddr.sin_port = htons(1119);
  multiaddr.sin_addr.s_addr =  makeaddr(225, 0, 0, 225);

  mreq.imr_multiaddr.s_addr = makeaddr(225, 0, 0, 225);
  mreq.imr_interface.s_addr = INADDR_ANY;

  multisock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if (multisock < 0) {
    perror("socket");
    exit(3);
  }

  setsockopt(multisock, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one));

  multiaddr.sin_addr.s_addr = INADDR_ANY;
  if (bind(multisock, multiaddr, sizeof(multiaddr))) {
    perror("bind");
    exit(3);
  }
  
  if (setsockopt(multisock, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq))) {
    perror("setsockopt IP_ADD_MEMBERSHIP");
    exit(3);
  }
  
  while (!(retstat = setsockopt(multisock, SOL_SOCKET, SO_RCVBUF,
		    (void *)&buflen, sizeof(buflen)))) {
    buflen -= 1024;
  }
  
  if (!retstat) {
      perror("m2n: setsockopt SO_RCVBUF");
    }

  multiaddr.sin_addr.s_addr =  makeaddr(225, 0, 0, 225);
  fromlen = sizeof(multiaddr);

  multidata = malloc(DLEN);
}

void *getmulti (lenptr)
int *lenptr;
{
  int len;

  len = recvfrom(multisock, multidata, DLEN, 0, &multiaddr, &fromlen);
  if (len < 0) {
    perror("recvfrom");
  }
  *lenptr = len;
  if (len < DLEN) {
    multidata[len] = '\0';
    /*   sfprintf(sferr,"%d bytes from %s:%d\n",
	 len, inet_ntoa(multiaddr.sin_addr),
	 multiaddr.sin_port);
	 fflush(stdout);
	 sfwrite(sfout, multidata, len); */
  } else {
    sfprintf(sferr, "Data too big\n");
    exit(5);
  }
  
  return (void *)multidata;
}
